%SerialLink.animate   Update a robot animation
%
% R.animate(q) updates an existing animation for the robot R.  This will have
% been created using R.plot().
%
% Updates graphical instances of this robot in all figures.
%
% Notes::
% - Called by plot() and plot3d() to actually move the arm models
% - Used for Simulink robot animation.
%
% See also SerialLink.plot.

% Copyright (C) 1993-2014, by Peter I. Corke
%
% This file is part of The Robotics Toolbox for MATLAB (RTB).
% 
% RTB is free software: you can redistribute it and/or modify
% it under the terms of the GNU Lesser General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% RTB is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Lesser General Public License for more details.
% 
% You should have received a copy of the GNU Leser General Public License
% along with RTB.  If not, see <http://www.gnu.org/licenses/>.
%
% http://www.petercorke.com

function animate(robot, qq)

    if nargin < 3
        handles = findobj('Tag', robot.name);
    end
    
    links = robot.links;
    N = robot.n;
    
    % get handle of any existing graphical robots of same name
    %  one may have just been created above
    handles = findobj('Tag', robot.name);
    
    % MAIN DISPLAY/ANIMATION LOOP
    while true
        % animate over all instances of this robot in different axes
        
        for q=qq'  % for all configurations in trajectory
            q = q';
            for handle=handles'
%                 h = get(handle, 'UserData');
%                 h.q = q';
%                 set(handle, 'UserData', h);
                
                  group = findobj('Tag', robot.name);
                  h = get(group, 'UserData');
                  
                % now draw it for a pose q
                if robot.mdh
                    % modified DH case
                    T = robot.base;
                    vert = transl(T)';
                    
                    for L=1:N
                        if robot.links(L).isprismatic()
                            set(h.pjoint(L), 'Matrix', trotz(q(L))*diag([1 1 -q(L) 1]));
                        end
                        T = T * links(L).A(q(L));
                        set(h.link(L), 'Matrix', T); 
                        vert = [vert; transl(T)'];
                    end
                    % update the transform for link N+1 (the tool)
                    T = T * robot.tool;
                    if length(h.link) > N
                        set(h.link(N+1), 'Matrix', T);
                    end
                    vert = [vert; transl(T)'];
                else
                    % standard DH case
                    T = robot.base;
                    vert = transl(T)';
					%-------------------------------------------------------------------------------------------%old version deleted
                    for L=1:N
                        % for all N+1 links
                        if robot.links(L).isprismatic()
			                 set(h.pjoint(L), 'Matrix', transl(0, 0, q(L)));		%~ NEW: update transform to visualize bar for a-parameter correctly				
			                 %set(h.pjointaxis(L), 'Matrix', diag([1 1 q(L) 1]));		%~ NEW: update transform to visualize bar for a-parameter correctly	
			                 if q(L)==0
									set(h.pjointaxis(L), 'Matrix', diag([1 1 0.0001 1]));		%~ NEW: update transform to visualize bar for a-parameter correctly	
							 else
									set(h.pjointaxis(L), 'Matrix', diag([1 1 q(L) 1]));		%~ NEW: update transform to visualize bar for a-parameter correctly	
							 end
			            else
			                 set(h.rjoint(L), 'Matrix', trotz(q(L)));               %~ NEW: update transform to visualize bar for a- and d-parameters correctly		
                        end
                        if h.link(L) ~= 0
                            set(h.link(L), 'Matrix', T);
                        end
                        
                        T = T * links(L).A(q(L));
                        
                        if L<N & h.coordsys(L) ~= 0						%~ NEW: update local coordinate systems
                            set(h.coordsys(L), 'Matrix', T);			%
                        end												%
                        
                        vert = [vert; transl(T)'];
                    end
                    % update the transform for link N+1 (the tool)
                    if length(h.link) > N
                        set(h.link(N+1), 'Matrix', T);
                    end
                    T = T*robot.tool;
                    vert = [vert; transl(T)'];
                end
																											
                if isfield(h, 'shadow')
                    set(h.shadow, 'Xdata', vert(:,1), 'Ydata', vert(:,2), ...
                        'Zdata', h.floorlevel*ones(size(vert(:,1))));
                end
                
                % update the tool tip trail
                if isfield(h, 'trail')
                    T = robot.fkine(q);
                    robot.trail = [robot.trail; transl(T)'];
                    set(h.trail, 'Xdata', robot.trail(:,1), 'Ydata', robot.trail(:,2), 'Zdata', robot.trail(:,3));
                end
                
%                 T = T * robot.tool;
%                 vert = [vert; transl(T)'];
                
                % animate the wrist frame
                if ~isempty(h.wrist)
                    trplot(h.wrist, T);
                end
                
                % add a frame to the movie
                if ~isempty(h.robot.framenum)
                    % write the frame to the movie folder
                    print( '-dpng', fullfile(h.robot.moviepath, sprintf('%04d.png', h.robot.framenum)) );
                    h.robot.framenum = h.robot.framenum+1;
                end
                
                if h.robot.delay > 0
                    pause(h.robot.delay);
                    drawnow
                end
            end
        end
        
        if ~h.robot.loop
            break;
        end        
    end
    
    h.q = q;
    set(group, 'UserData', h);
